/*----------------------------------------------------------------------------
 ADOL-C -- Automatic Differentiation by Overloading in C++
 File:     hos_forward.c
 Revision: $Id: hos_forward.c 42 2009-07-15 18:37:17Z awalther $
 Contents: hos_forward (higher-order-scalar forward mode)
 
 Copyright (c) Andrea Walther, Andreas Griewank, Andreas Kowarz, 
               Hristo Mitev, Sebastian Schlenkrich, Jean Utke, Olaf Vogel
  
 This file is part of ADOL-C. This software is provided as open source.
 Any use, reproduction, or distribution of the software constitutes 
 recipient's acceptance of the terms of the accompanying license file.
 
----------------------------------------------------------------------------*/
#define _HOS_  1
#define _KEEP_ 1
#include <uni5_for.c>
#undef _KEEP_
#undef _HOS_

