/*----------------------------------------------------------------------------
 ADOL-C -- Automatic Differentiation by Overloading in C++
 File:     int_forward_t.c
 Revision: $Id: indopro_forward_t.c 42 2009-07-15 18:37:17Z awalther $
 Contents: int_forward (integer forward mode for bit pattern propagation)

 Copyright (c) Andrea Walther
  
 This file is part of ADOL-C. This software is provided as open source.
 Any use, reproduction, or distribution of the software constitutes 
 recipient's acceptance of the terms of the accompanying license file.
 
----------------------------------------------------------------------------*/
#define _INDO_ 1
#define _INDOPRO_ 1
#define _TIGHT_ 1
#include  <uni5_for.c>
#undef _INDO_
#undef _INDOPRO_
#undef _TIGHT_

