/*----------------------------------------------------------------------------
 ADOL-C -- Automatic Differentiation by Overloading in C++
 File:     adutils.h
 Revision: $Id: adutils.h 354 2012-10-01 11:32:26Z kulshres $
 Contents: Provides all C/C++ interfaces of ADOL-C.
           NOTICE: This file is kept for compatibility reasons only. 
                   The new header adolc.h is included.
 
 Copyright (c) Andreas Kowarz
  
 This file is part of ADOL-C. This software is provided as open source.
 Any use, reproduction, or distribution of the software constitutes 
 recipient's acceptance of the terms of the accompanying license file.
 
----------------------------------------------------------------------------*/

#if !defined(ADOLC_ADUTILS_H)
#define ADOLC_ADUTILS_H 1

#include <adolc/adolc.h>

#endif

